#ifndef __IMAGE_FILE_FORMAT_H__
#define __IMAGE_FILE_FORMAT_H__

#include <string>
#include <vector>
#include <map>
#include <assert.h>

using namespace std;
// !!! It would be nice to have a description of the ImageFileFormat class at top of page

/** These are the NUM_OF_ELEMENTS elements of which any possible image filename is composed */
enum ELEMENT
{
  EXPT,             ///< Experiment name
  TUBE,             ///< Tube identifier, usually a number
  WINDOW,           ///< Window number
  SESSION,          ///< Session number
  SDATE,            ///< Date
  TIME,             ///< Time
  INITIALS,         ///< Initials of data gatherer
  MISC1,            ///< Skips unnecessary element in filename
  MISC2,            ///< Skips unnecessary element in filename
  MISC3,            ///< Skips unnecessary element in filename
  MISC4,            ///< Skips unnecessary element in filename
  NUM_OF_ELEMENTS   ///< Total number of elements that may be found
};

// used to help with ordering the date elements
enum DATE_ORDER
{
  YEAR,
    MONTH,
    DAY
};

/** contains information about a specific element in the image filename */
struct OrderElement
{
  ELEMENT element;  ///< integer indicating which element
  CString prefix;   ///< prefix (if it exists) in the actual image filename; the prefix is thrown away upon parsing
  int length;       ///< maximum number of characters in the image filename devoted to this element -- used with tube, window, session
  bool included;    ///< true if element is included in image filename
  
  bool operator==(const OrderElement& other) const
  {
    if(element != other.element) return false;
    if(prefix != other.prefix) return false;
    if(length != other.length) return false;
    if(included != other.included) return false;
    return true;
  }
  /** @returns a human-readable form of ELEMENT */
  CString GetName() const
  {
    CString name;
    switch(element)
    {
    case EXPT : name = "Experiment"; break;
    case TUBE : name = "Tube"; break;
    case WINDOW : name = "Window"; break;
    case SESSION : name = "Session"; break;
    case SDATE : name = "Date"; break;
    case TIME : name = "Time"; break;
    case INITIALS : name = "User Initials"; break;
    case MISC1 : name = "Misc 1"; break;
    case MISC2 : name = "Misc 2"; break;
    case MISC3 : name = "Misc 3"; break;
    case MISC4 : name = "Misc 4"; break;
    default:  assert(0);  name = "[error]"; break;
    }
    return name;
  }
};

/** contains the format of the date */
struct Date_Format
{  
  std::vector<int> order; // used to determine the order
  bool four_year;
  CString date_delim;
  
  bool operator==(const Date_Format& other) const
  {
    int i;
    for(i = 0; i < order.size(); i++)
    {
      if(order[i] != other.order[i]) return false;
    }
    if(four_year != other.four_year) return false;
    if(date_delim != other.date_delim) return false;
    return true;
  }
};

/** contains information about the image file format used in the current directory
E.g., order of the items, delimeter between items, etc.
*/ 
struct ImageFileFormatInfo
{
  std::vector<OrderElement> InfoOrder;  ///< order of the elements in the filename
  ///  vector always includes NUM_OF_ELEMENTS
  int numElementsInFileName;       ///< number of elements included in the filename
  CString filetype;                ///< image type, e.g., jpg or bmp
  char delim;                      ///< delimiter between elements in the filename
  Date_Format m_date_format;
  bool operator==(const ImageFileFormatInfo& other) const
  {
    int i;
    for(i = 0; i <InfoOrder.size(); i++)
    {
      if( !(InfoOrder[i] == other.InfoOrder[i]) ) return false;
    }
    if(numElementsInFileName != other.numElementsInFileName) return false;
    if(filetype != other.filetype) return false;
    if(delim != other.delim) return false;
    if( !(m_date_format == other.m_date_format) ) return false;
    return true;
  }
};

/** contains the date */
struct Date
{
  Date() : year(-1), month(-1), day(-1) {}

  // Note: these integers are treated as 1-indexed numbers
  int year;   ///< year
  int month;  ///< month
  int day;    ///< day
  
  bool operator==(const Date& other) const
  {
    return (year == other.year && month == other.month && day == other.day);
  }  
};

/** contains information found in the image file name */
struct ImageInfo
{
  CString expt;      ///< Experiment name
  CString tube;      ///< Tube
  int window;        ///< Window number
  int session;       ///< Session number
  Date date;         ///< Date that the image was taken - contains day, month, year
  CString initials;  ///< Initials of data gatherer (Bartz format)
};

class ImageProgressBar
{
public:
  ImageProgressBar(const CString &, int, CWnd *);
  ~ImageProgressBar();
  void StepIt(bool);
private:
  CProgressCtrl m_progress_bar;
  int m_lower, m_upper;
  CRect m_rect[3];
  CClientDC *m_dc_ptr;
};

/** Stores image file format, and provides method to retrieve file name. 
*/

class ImageFileFormat
{
public:
  ImageFileFormat();
  virtual ~ImageFileFormat() {}
  bool operator==(const ImageFileFormat& ) const;
  
  // Call this whenever the image file directory changes.
  void SetFileDirectory(const CString& directory);

  // !!! Note:  Now that we added SetFileDirectory, it does not make sense for the functions below 
  // to take directory as an input parameter. -- STB 4/27/10
  
  /** Generates a filename given the parameters, then searches for the file in
  the given directory.
  @param directory Directory containing the images
  @param tube Tube # of the image
  @param window Window # of the image
  @param session Session # of the image
  @returns filename of the corresponding file or an empty string if the file
  is not found
  */
  
  CString GetImageFileName(const CString& directory, const CString& tube, int window, int session) const;
  
  /** Searches a directory for a file that matches the given image filename format
  @param directory The directory to search
  @returns true if file is found; false otherwise
  */
  bool ImageFileExistsInDirectory(const CString& directory) const;
  
  int CountFilesInDirectory(const CString& directory) const;
  
  /* Gets image info from a filename
  @param filename The filename from which the image info should be obtained
  @returns a struct containing the image information
  */
  ImageInfo GetImageInfo(const CString& filename) const;
  
  /* Gets image info of the first valid file found in a directory
  @param directory The image directory
  @returns a struct containing the image information of the first file
  */
  ImageInfo GetImageInfoOfFirstFile(const CString& directory) const;
  
  /* Sets the image file format info given a const reference */
  void SetImageFileFormatInfo(const ImageFileFormatInfo&);
  
  /* @returns a const reference to an imagefileformatinfo struct */
  const ImageFileFormatInfo& GetImageFileFormatInfo() const;
  
  /* @returns true if date information is available in file name */
  bool DateIsIncluded() const;  
  
  /* Sets the format to the Bartz format */
  void SetToBartz();
  
  /* Sets the forat for the date to an default state */
  void InitializeDateFormat();
  
  /* Writes string containing image file format to a given file */
  void Save(FILE* fp) const;
  
  /* Loads image file format from a given file */
  bool Load(FILE* fp);
  
  /* Gets the first image in the directory */
  CString GetFirstImage(const CString& directory) const;
  
  /* Gets the rectangle that is the same size as the image */
  static CRect GetImageRectangle(const CString& filename);
  
  // gets dates from dir. date must be included in filename also includes a progress bar
  vector<Date> GetDatesFromDir(const CString& dir, int max_sessions, CWnd *pParentWnd) const;
  
  // searches through all the images in the directory, parses the filenames, and sets the min/max window/session
  // if pParentWnd is Null then no progress bar is displayed
  bool ImageFileFormat::CatalogueImages(const CString& directory, int* max_window, int* min_window,
    int* max_session, int* min_session, CWnd *pParentWnd, class Controller &controller) const;
  
private:
  ImageFileFormatInfo m_image_file_format_info;
  
  /* @returns a search string created from the image file format */
  CString GetFileSearchString() const;
  
  /* Sets the lengths for the elements included in the file name
  @returns false if error, true otherwise
  */
  bool SetLengths(const CString& file_directory);

  // key value tuple for mapping the name of the tube, window number,
  // and session number to the filename of an image
  struct Key
  {
    CString tubename;
    int window;
    int session;

    Key(const CString t, const int w, const int s)
    : tubename(t),
      window(w), 
      session(s) 
    {}

    // Compare first by tube name, then by window and then by session
    // If tubename, window, and session are all equal the map pair is 
    // the same and either can be used. This is because the map does 
    // not allow multiple copies of the same key.
    bool operator<(const Key& other) const
    {
      if (tubename != other.tubename) return tubename < other.tubename;
      if (window != other.window) return window < other.window;
      if (session != other.session) return session < other.session;
      return false;
    }
  };

  // a map from a tubename, window and session to the image filename
  typedef std::map<Key, CString> ImageFileMap;
  typedef std::pair<Key, CString> ImageFilePair;
  typedef std::map<Key, CString>::iterator ImageFileIterator;

  // !!! Refactor: Probably should move m_map to somewhere else because it is not stored in the rfy file
  ImageFileMap m_map;

  // !!! This is the old version of the function. There may be a way to remove it completely,
  //    but its not hurting anything for now (provides a safety net for filenames not in map now)
  CString GetFileName(const CString& directory, const CString& tube, int window, int session) const;
};

#endif __IMAGE_FILE_FORMAT_H__