// RootflyGraphics.cpp: implementation of the RootflyGraphics class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "rootfly.h"
#include "RootflyGraphics.h"
#include "DataStructures.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

static const int WID_SEL = 2;  // line width of root (selected)
static const int WID_DEF = 1;  // line width of root (default)

COLORREF ROOTFLY_RGB_RED =         RGB(255,0,0);
COLORREF ROOTFLY_RGB_GREEN =       RGB(0,255,0);
COLORREF ROOTFLY_RGB_BLUE =        RGB(0,0,255);
COLORREF ROOTFLY_RGB_BLACK =       RGB(0,0,0);
COLORREF ROOTFLY_RGB_DARK_BROWN =  RGB(139, 69, 19 );
COLORREF ROOTFLY_RGB_LIGHT_BROWN = RGB(205, 133, 63);
COLORREF ROOTFLY_RGB_WHITE =       RGB(255, 255, 255);
COLORREF ROOTFLY_RGB_SYS_GRAY =    ::GetSysColor(COLOR_3DFACE);

// Note:  Roots have several drawing modes:
//
//                            width     color    draw points?   ID outline?
//   Normal (live)            WID_DEF     NL         no             no
//   Normal (dead)            WID_DEF     ND         no             no
//   Mouse over (live)        WID_DEF     SL         no             no
//   Mouse over (dead)        WID_DEF     SD         no             no
//   Selected   (live)        WID_SEL     SL         yes            yes
//   Selected   (dead)        WID_SEL     SD         yes            yes
//   New root being drawn     WID_DEF     SL         no             no
//

const int RootflyGraphics::m_point_size = 6;

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

RootflyGraphics::RootflyGraphics()
: rfy_dlg_conv(NULL),
  rfy_dlg_settings(NULL),
  rfy_dlg_data(NULL)
{

}

RootflyGraphics::~RootflyGraphics()
{

}

void RootflyGraphics::Initialize( const ImageClientConverter &icc, const Settings &s, const RootflyData &rfd )
{
  rfy_dlg_conv = &icc;
  rfy_dlg_settings = &s;
  rfy_dlg_data = &rfd;
}

COLORREF RootflyGraphics::GetRootColor(const RootInstance& inst, bool selected) const
{
  if (selected)
  {
    if (inst.IsAlive())  return rfy_dlg_settings->GetRootDrawingColorLiveSelected();
    else                 return rfy_dlg_settings->GetRootDrawingColorDeadSelected();
  }
  else
  {
    if (inst.IsAlive())  return rfy_dlg_settings->GetRootDrawingColorLive();
    else                 return rfy_dlg_settings->GetRootDrawingColorDead();
  }  
}

int RootflyGraphics::GetRootLineDrawingWidth( bool selected ) const
{
  return selected ? WID_SEL : WID_DEF;
}

// draws the diameter
// 'selected':  whether to draw root with 'selected' color
// x_offset is for 3x3 zoomed-out display -- MTB
void RootflyGraphics::DrawDiameter(const RootInstance& inst, COLORREF color, bool draw_point, CDC& dc, const int x_offset) const
{
  const RDiameter *diam = inst.GetDiameter(0);
  RPoint dpoint = diam->r_point;
  CPoint cpt1,cpt2;
  double rad = diam->radius;
  cpt1 = rfy_dlg_conv->ImageToClient( CPoint(dpoint.x - rad, dpoint.y - rad), dpoint.window );
  cpt2 = rfy_dlg_conv->ImageToClient( CPoint(dpoint.x + rad, dpoint.y + rad), dpoint.window );
  CRect rect(cpt1.x+x_offset, cpt1.y, cpt2.x+x_offset, cpt2.y);
  
  CPen pen(PS_SOLID, 1, color);
  CPen* oldpen = dc.SelectObject(&pen);
  CBrush brush;
  brush.CreateSolidBrush( RootColorToRgb( inst.GetColor() ) );
  dc.SelectObject(brush);   
  dc.Ellipse( rect );
  dc.SelectObject(oldpen);  

  if (draw_point) 
  { // draw extra point for resizing
    const RPoint point = RPoint(dpoint.x, dpoint.y - rad, dpoint.window);
    DrawPoint(point, dc, x_offset);
  }
}

void RootflyGraphics::DrawRootId(const Root& root, 
                                 bool selected, 
                                 int current_window, 
                                 int session, 
                                 CDC& dc,
                                 const int x_offset) const
{
  // displays the root id, checks that the current window is the same as the primary window for the RootInstance, if it is not
  // then it will show the RootInstance as "WindowID(PrimaryWindow
  int window = root.GetPrimaryWindow();
  CString fmt = "%d";
  if (window != current_window) fmt += "*";
  CString str;
  str.Format(fmt, root.GetDisplayableID());

  const RootInstance *inst = root.GetRootInstance( session );
  if (!inst)  return;
  CRect main = rfy_dlg_conv->GetClientRect( window );
  const std::vector<RPoint>* line = inst->GetAllPoints();
  if (!line)  return;
  if (line->size() > 0)
  {
    CPoint conv_point = rfy_dlg_conv->ImageToClient( line->at(0) );
    CPoint topleft(conv_point.x-15+x_offset, conv_point.y-15);
    //if (topleft.x < main.left) topleft.x = main.left;
    //if (topleft.y < main.top ) topleft.y = main.top;
    dc.SetTextColor(ROOTFLY_RGB_BLACK);
    dc.TextOut(topleft.x, topleft.y, str);
    if (selected)
    { // draw outline around root id
      CSize sz = dc.GetTextExtent(str);
      COLORREF color = GetRootColor(*inst, selected);
      CBrush brush( color );
      CRect rect(topleft.x, topleft.y, topleft.x + sz.cx, topleft.y + sz.cy);
      rect.InflateRect(1, 1, 1, 1);
      dc.FrameRect( rect, &brush );
    }
  }
}

// RootColorToRgb function
COLORREF RootflyGraphics::RootColorToRgb( RootInstance::Color tempcolor ) const
{
  switch (tempcolor)
  {
  case RootInstance::COLOR_WHITE:       return ROOTFLY_RGB_WHITE;
  case RootInstance::COLOR_LIGHTBROWN:  return ROOTFLY_RGB_LIGHT_BROWN;
  case RootInstance::COLOR_DARKBROWN:   return ROOTFLY_RGB_DARK_BROWN;
  case RootInstance::COLOR_BLACK:       return ROOTFLY_RGB_BLACK;
  default:
    assert(0);  // root should have a color
    return ROOTFLY_RGB_WHITE;
  }
}

// draws points on the selected root and diameter
void RootflyGraphics::DrawPoint(const RPoint& point, CDC& dc, int offset) const
{
  std::vector<RPoint> line;

  line.push_back( point );

  DrawPoints( line, dc , offset);
}

// draws points on the selected root
void RootflyGraphics::DrawPoints(const std::vector<RPoint>& line, CDC& dc, int offset) const
{
  int size = m_point_size / 2;    // halfsize
  
  CPen pen(PS_SOLID, 1, ROOTFLY_RGB_BLACK);
  CPen* oldpen = dc.SelectObject(&pen);
  CBrush brush;
  brush.CreateSolidBrush(ROOTFLY_RGB_WHITE);
  CBrush* oldbrush = dc.SelectObject(&brush);
  std::vector<RPoint>::const_iterator pt;
  for (pt = line.begin() ; pt != line.end() ; pt++)
  {
    CPoint cpt = rfy_dlg_conv->ImageToClient( *pt );
    CRect rect(cpt.x - size + offset, cpt.y - size, cpt.x + size + offset, cpt.y + size);
    dc.Rectangle(rect);
  }
  dc.SelectObject(oldpen);
  dc.SelectObject(oldbrush);
}

void RootflyGraphics::DrawPolyline( const std::vector<RPoint>& line, COLORREF c, int width, CDC& dc, const int x_offset ) const
{	 
  if ( line.size() >= 2 )
  {
    CPen pen(PS_SOLID, width, c );
    CPen* oldpen = dc.SelectObject(&pen);
    vector<RPoint>::const_iterator p = line.begin();
    CPoint pt;
    
    pt = rfy_dlg_conv->ImageToClient( *p );
    dc.MoveTo(pt.x + x_offset, pt.y);
    for ( ; p != line.end() ; p++)
    {
      pt = rfy_dlg_conv->ImageToClient( *p );
      dc.LineTo(pt.x + x_offset, pt.y);
    } 
    
    dc.SelectObject(oldpen);
  }  
}

void RootflyGraphics::ExtendPolyline( const RootInstance& ri, 
                                      bool extend_from_front,
                                      COLORREF color,
                                      int width,
                                      CPoint mouse_point,
                                      CDC& dc,
                                      const int x_offset) const
{
  const std::vector<RPoint>* line = ri.GetAllPoints();
  if (line && line->size() > 0)
  {
    mouse_point.x -= x_offset;
    std::vector<RPoint> xtend;
    xtend.push_back( extend_from_front ? (*line)[0] : line->back() );
    xtend.push_back( rfy_dlg_conv->ClientToImage( mouse_point ) );
    DrawPolyline(xtend, color, width, dc, x_offset);
  }
}

// Draw a single root
// (Code should eventually be refactored to use this function.
//  First, though, we may need to make slight modifications to the function to ensure
//  that its functionality is as desired.)  -- STB 4/21/08
void RootflyGraphics::DrawSingleRoot(CDC& dc, 
                                     const Root &root, 
                                     int window,
                                     int session,
                                     bool extend_root,
                                     bool extend_direction,
                                     const CPoint& mouse_point,
                                     bool selected,
                                     bool show_diam,
                                     bool show_points,  
                                     bool show_id ) const
{
  const RootInstance* ri = root.GetRootInstance( session );

  // Offset drawn root into correct column
  const ApplicationData& app_data = rfy_dlg_data->GetApplicationData();
  int offset = (session - app_data.GetCurrentSession()) * rfy_dlg_conv->GetClientRect( window ).Width();

  if (ri && app_data.ImageExists( app_data.GetCurrentTubeName(), window, session ) )
  {
    COLORREF color = GetRootColor(*ri, selected);
    int width = GetRootLineDrawingWidth(selected);
    DrawPolyline( *ri->GetAllPoints(), color, width, dc, offset );
    if (show_points) DrawPoints( *ri->GetAllPoints(), dc, offset);
    if (show_diam) DrawDiameter( *ri, color, show_points, dc, offset );
    if (show_id)  DrawRootId(root, selected, window, session, dc, offset);
    if (extend_root) ExtendPolyline( *ri, extend_direction, color, width, mouse_point, dc, offset);
  }
}

// Draws the selected root
void RootflyGraphics::DrawSelectedRoot(CDC& dc, int selected_id, int point_session, int mode, const CPoint& mouse_point, bool xtend_tail, int x_offset) const
{
  const ApplicationData& app_data = rfy_dlg_data->GetApplicationData();
  const Experiment& exp = rfy_dlg_data->GetExperiment();
  const Tube *tube = exp.GetTubeByName(app_data.GetCurrentTubeName());
  if (!tube)  return;
  const Root *root = tube->GetRoot( selected_id );
  if (!root)  return;
  const RootInstance *inst = root->GetRootInstance(point_session);
  if (!inst)  return;
  const std::vector<RPoint> *line = inst->GetAllPoints();
  if (!line)  return;
  const RDiameter *diam = inst->GetDiameter(0);
  if (!diam)  return;
 
  bool selected = true;
  COLORREF color = GetRootColor(*inst, selected);
  bool show_id = rfy_dlg_settings->GetShowRootID();
  bool show_diam = rfy_dlg_settings->GetShowDiameters();
  bool show_points = ( mode == MODE_ROOTSELECTED );
  bool extend_root = ( mode == MODE_EXTENDINGENDPOINT );

  DrawSingleRoot( dc, *root, app_data.GetCurrentWindow(), point_session, extend_root, !xtend_tail, mouse_point, selected, show_diam, show_points, show_id);
}


// Draws all the roots in the data structure
void RootflyGraphics::DrawAllRoots(CDC& dc, int mode, bool zoomed) const
{
  if ( mode == MODE_NOFILELOADED ) return;  
  if ( mode == MODE_NOIMAGEDISPLAYED ) return;  
  const ApplicationData& app_data = rfy_dlg_data->GetApplicationData();
  const Experiment& exp = rfy_dlg_data->GetExperiment();
  const Tube *tube = exp.GetTubeByName(app_data.GetCurrentTubeName());
  if (!tube)  return;
  const vector<Root> *roots = tube->GetAllRoots();
  if (!roots)  return;
  
  
  // Draws roots
  int current_window = app_data.GetCurrentWindow();
  int current_session = app_data.GetCurrentSession();
  for ( int s = current_session-1; s <= current_session+1; s++ ) {

    // Set clip region to current column
    CRect clip_rect = rfy_dlg_conv->GetClientRect( current_window );
    clip_rect.InflateRect( 0, clip_rect.Height() );
    clip_rect.OffsetRect( (s - current_session) * clip_rect.Width(), 0 );
    SetClipRegion( dc, clip_rect ); 

    for ( int w = current_window-1; w <= current_window+1; w++ ) {
      for ( int i = 0; i < roots->size(); i++ ) {
        if ( roots->at(i).GetPrimaryWindow() == w ) {
          const RootInstance *inst = roots->at(i).GetRootInstance(s);
          if (!inst)  continue; 

          if (inst->IsAlive() || rfy_dlg_settings->GetShowDeadRoots() )
          {
            bool extend_root = false;
            bool extend_direction = false;
            CPoint mouse_point(0,0);
            bool selected = ( mode == MODE_STARTINGMOVEALL || mode == MODE_MOVINGALLROOTS );
            bool show_diam = rfy_dlg_settings->GetShowDiameters();
            bool show_points = false;
            bool show_id = rfy_dlg_settings->GetShowRootID();

            DrawSingleRoot( dc, roots->at(i), w, s, extend_root, extend_direction, mouse_point, selected, show_diam, show_points, show_id);
           }
        }
      }
    }
  }
  SetClipRegion( dc, true );
}

// Draws the new root that is currently being created.
//    x_offset is used for 3x3 drawing -- MTB
//    (Note that we cannot use DrawSingleRoot for this, because RootflyDlg only creates a 
//     RootInstance, not a Root, while the user creates a new root.)  -- ARies
void RootflyGraphics::DrawNewRoot(CDC& dc, int mode, const RootInstance &ri, const CPoint& mouse_point, const int x_offset) const
{
  const std::vector<RPoint>* line = ri.GetAllPoints();
  COLORREF color = GetRootColor(ri, true);

  if ( mode == MODE_DRAWINGNEWROOT)
  {	
    DrawPolyline(*line, color, WID_DEF, dc, x_offset);    
    ExtendPolyline(ri, false, color, WID_DEF, mouse_point, dc, x_offset);
  }  
  
  //Draw the diameter that is being created
  else if ( mode == MODE_SIZINGDIAMETER || mode == MODE_LOCATINGDIAMETER  )
  {
    DrawPolyline(*line, color, WID_DEF, dc, x_offset);
    DrawDiameter(ri, GetRootColor(ri, true), false, dc, x_offset);    
  }
  
}


// Sets the clip region of the DC to either 
//    - the main image (in regular mode), or
//    - the union of the three images (in zoomed out mode).
// Should only need to call this function once before starting all the drawing.
// Note:  SelectClipRgn only seems to work for the real CDC, not for the MemDC.
void RootflyGraphics::SetClipRegion(CDC& dc, bool zoomed) const
{
  CRgn clip_rgn;
  const ApplicationData& app_data = rfy_dlg_data->GetApplicationData();
  CRect rect;
  int window = app_data.GetCurrentWindow();

  rect = rfy_dlg_conv->GetClientRect( window );  
  if ( zoomed ) rect.InflateRect( rect.Width(), rect.Height() );

  //  clip region
  clip_rgn.CreateRectRgnIndirect(&rect);
  dc.SelectClipRgn(&clip_rgn); 
}

void RootflyGraphics::SetClipRegion(CDC& dc, const CRect& rect) const
{
  CRgn clip_rgn;
  clip_rgn.CreateRectRgnIndirect(&rect);
  dc.SelectClipRgn(&clip_rgn); 
}

void RootflyGraphics::FillImageDisplayAreaWithGray(CDC& dc, const CRect &rect) const
{
  // fill image area with gray
  CBrush background(ROOTFLY_RGB_SYS_GRAY);
  dc.FillRect(rect, &background);
}

bool RootflyGraphics::IsRootInCurrentWindow( const RootInstance &inst ) const
{
  const vector<RPoint> *points = inst.GetAllPoints();
  const ApplicationData& app_data = rfy_dlg_data->GetApplicationData();
  for ( int i = 0; i < points->size() -1 ; i++ )
  {
    if ( points->at(i).window == app_data.GetCurrentWindow() ) return true;
  }
  return false;
}